@@ -18,17 +18,18 @@
 
 #pragma once
 
+#include <type_traits>
+
 #include "caf/all.hpp"
 #include "caf/config.hpp"
+#include "caf/detail/gcd.hpp"
 #include "caf/meta/annotation.hpp"
 #include "caf/test/unit_test.hpp"
 
-#include "caf/detail/gcd.hpp"
-
 CAF_PUSH_WARNINGS
 
 /// The type of `_`.
-struct wildcard { };
+struct wildcard {};
 
 /// Allows ignoring individual messages elements in `expect` clauses, e.g.
 /// `expect((int, int), from(foo).to(bar).with(1, _))`.
@@ -75,7 +76,7 @@ bool operator==(const message& x, const T& y) {
 } // namespace caf
 
 // dummy function to force ADL later on
-//int inspect(int, int);
+// int inspect(int, int);
 
 template <class T>
 struct has_outer_type {
@@ -105,7 +106,7 @@ public:
   template <size_t X>
   using pos = std::integral_constant<size_t, X>;
 
-  elementwise_compare_inspector(const Tup& xs) : xs_(xs) {
+  explicit elementwise_compare_inspector(const Tup& xs) : xs_(xs) {
     // nop
   }
 
@@ -122,19 +123,13 @@ private:
   }
 
   template <size_t X, class T, class... Ts>
-  typename std::enable_if<
-    caf::meta::is_annotation<T>::value,
-    bool
-  >::type
+  typename std::enable_if<caf::meta::is_annotation<T>::value, bool>::type
   iterate(pos<X> pos, const T&, const Ts&... ys) {
     return iterate(pos, ys...);
   }
 
   template <size_t X, class T, class... Ts>
-  typename std::enable_if<
-    !caf::meta::is_annotation<T>::value,
-    bool
-  >::type
+  typename std::enable_if<!caf::meta::is_annotation<T>::value, bool>::type
   iterate(pos<X>, const T& y, const Ts&... ys) {
     std::integral_constant<size_t, X + 1> next;
     check(y, get<X>(xs_));
@@ -162,18 +157,34 @@ class caf_handle : caf::detail::comparable<caf_handle>,
 public:
   using pointer = caf::abstract_actor*;
 
-  constexpr caf_handle() : ptr_(nullptr) {
+  constexpr caf_handle(pointer ptr = nullptr) : ptr_(ptr) {
     // nop
   }
 
-  template <class T>
-  caf_handle(const T& x) {
-    *this = x;
+  caf_handle(const caf::strong_actor_ptr& x) {
+    set(x);
+  }
+
+  caf_handle(const caf::actor& x) {
+    set(x);
+  }
+
+  caf_handle(const caf::actor_addr& x) {
+    set(x);
+  }
+
+  caf_handle(const caf::scoped_actor& x) {
+    set(x);
+  }
+
+  template <class... Ts>
+  caf_handle(const caf::typed_actor<Ts...>& x) {
+    set(x);
   }
 
   caf_handle(const caf_handle&) = default;
 
-  inline caf_handle& operator=(caf::abstract_actor* x)  {
+  caf_handle& operator=(pointer x) {
     ptr_ = x;
     return *this;
   }
@@ -181,7 +192,7 @@ public:
   template <class T,
             class E = caf::detail::enable_if_t<!std::is_pointer<T>::value>>
   caf_handle& operator=(const T& x) {
-    ptr_ = caf::actor_cast<pointer>(x);
+    set(x);
     return *this;
   }
 
@@ -191,6 +202,10 @@ public:
     return ptr_;
   }
 
+  pointer operator->() const {
+    return get();
+  }
+
   ptrdiff_t compare(const caf_handle& other) const {
     return reinterpret_cast<ptrdiff_t>(ptr_)
            - reinterpret_cast<ptrdiff_t>(other.ptr_);
@@ -201,35 +216,20 @@ public:
   }
 
 private:
+  template <class T>
+  void set(const T& x) {
+    ptr_ = caf::actor_cast<pointer>(x);
+  }
+
   caf::abstract_actor* ptr_;
 };
 
-// -- access to an actor's mailbox ---------------------------------------------
-
-/// Returns a pointer to the next element in an actor's mailbox without taking
-/// it out of the mailbox.
-/// @pre `ptr` is alive and either a `scheduled_actor` or `blocking_actor`
-inline caf::mailbox_element* next_mailbox_element(caf_handle x) {
-  CAF_ASSERT(x.get() != nullptr);
-  auto sptr = dynamic_cast<caf::scheduled_actor*>(x.get());
-  if (sptr != nullptr) {
-    return sptr->mailbox().closed() || sptr->mailbox().blocked()
-           ? nullptr
-           : sptr->mailbox().peek();
-  }
-  auto bptr = dynamic_cast<caf::blocking_actor*>(x.get());
-  CAF_ASSERT(bptr != nullptr);
-  return bptr->mailbox().closed() || bptr->mailbox().blocked()
-         ? nullptr
-         : bptr->mailbox().peek();
-}
-
 // -- introspection of the next mailbox element --------------------------------
 
 /// @private
 template <class... Ts>
 caf::optional<std::tuple<Ts...>> default_extract(caf_handle x) {
-  auto ptr = next_mailbox_element(x);
+  auto ptr = x->peek_at_next_mailbox_element();
   if (ptr == nullptr || !ptr->content().template match_elements<Ts...>())
     return caf::none;
   return ptr->content().template get_as_tuple<Ts...>();
@@ -277,11 +277,11 @@ template <class T, class... Ts>
 std::tuple<T, Ts...> extract(caf_handle x) {
   auto result = try_extract<T, Ts...>(x);
   if (result == caf::none) {
-    auto ptr = next_mailbox_element(x);
+    auto ptr = x->peek_at_next_mailbox_element();
     if (ptr == nullptr)
       CAF_FAIL("Mailbox is empty");
-    CAF_FAIL("Message does not match expected pattern: "
-             << to_string(ptr->content()));
+    CAF_FAIL(
+      "Message does not match expected pattern: " << to_string(ptr->content()));
   }
   return std::move(*result);
 }
@@ -294,9 +294,8 @@ bool received(caf_handle x) {
 template <class... Ts>
 class expect_clause {
 public:
-  expect_clause(caf::scheduler::test_coordinator& sched)
-      : sched_(sched),
-        dest_(nullptr) {
+  explicit expect_clause(caf::scheduler::test_coordinator& sched)
+    : sched_(sched), dest_(nullptr) {
     peek_ = [=] {
       /// The extractor will call CAF_FAIL on a type mismatch, essentially
       /// performing a type check when ignoring the result.
@@ -326,8 +325,8 @@ public:
   template <class Handle>
   expect_clause& to(const Handle& whom) {
     CAF_REQUIRE(sched_.prioritize(whom));
-    dest_ = &sched_.next_job<caf::scheduled_actor>();
-    auto ptr = next_mailbox_element(dest_);
+    dest_ = &sched_.next_job<caf::abstract_actor>();
+    auto ptr = dest_->peek_at_next_mailbox_element();
     CAF_REQUIRE(ptr != nullptr);
     if (src_)
       CAF_REQUIRE_EQUAL(ptr->sender, src_);
@@ -335,7 +334,7 @@ public:
   }
 
   expect_clause& to(const caf::scoped_actor& whom) {
-    dest_ = whom.ptr();
+    dest_ = caf::actor_cast<caf::abstract_actor*>(whom);
     return *this;
   }
 
@@ -344,7 +343,7 @@ public:
     // TODO: replace this workaround with the make_tuple() line when dropping
     //       support for GCC 4.8.
     std::tuple<typename std::decay<Us>::type...> tmp{std::forward<Us>(xs)...};
-    //auto tmp = std::make_tuple(std::forward<Us>(xs)...);
+    // auto tmp = std::make_tuple(std::forward<Us>(xs)...);
     // TODO: move tmp into lambda when switching to C++14
     peek_ = [=] {
       using namespace caf::detail;
@@ -366,16 +365,134 @@ protected:
 
   caf::scheduler::test_coordinator& sched_;
   caf::strong_actor_ptr src_;
-  caf::local_actor* dest_;
-  std::function<void ()> peek_;
+  caf::abstract_actor* dest_;
+  std::function<void()> peek_;
+};
+
+template <>
+class expect_clause<void> {
+public:
+  explicit expect_clause(caf::scheduler::test_coordinator& sched)
+    : sched_(sched), dest_(nullptr) {
+    // nop
+  }
+
+  expect_clause(expect_clause&& other) = default;
+
+  ~expect_clause() {
+    auto ptr = dest_->peek_at_next_mailbox_element();
+    if (ptr == nullptr)
+      CAF_FAIL("no message found");
+    if (!ptr->content().empty())
+      CAF_FAIL("non-empty message found: " << to_string(ptr->content()));
+    run_once();
+  }
+
+  expect_clause& from(const wildcard&) {
+    return *this;
+  }
+
+  template <class Handle>
+  expect_clause& from(const Handle& whom) {
+    src_ = caf::actor_cast<caf::strong_actor_ptr>(whom);
+    return *this;
+  }
+
+  template <class Handle>
+  expect_clause& to(const Handle& whom) {
+    CAF_REQUIRE(sched_.prioritize(whom));
+    dest_ = &sched_.next_job<caf::abstract_actor>();
+    auto ptr = dest_->peek_at_next_mailbox_element();
+    CAF_REQUIRE(ptr != nullptr);
+    if (src_)
+      CAF_REQUIRE_EQUAL(ptr->sender, src_);
+    return *this;
+  }
+
+  expect_clause& to(const caf::scoped_actor& whom) {
+    dest_ = caf::actor_cast<caf::abstract_actor*>(whom);
+    return *this;
+  }
+
+protected:
+  void run_once() {
+    auto dptr = dynamic_cast<caf::blocking_actor*>(dest_);
+    if (dptr == nullptr)
+      sched_.run_once();
+    else
+      dptr->dequeue(); // Drop message.
+  }
+
+  caf::scheduler::test_coordinator& sched_;
+  caf::strong_actor_ptr src_;
+  caf::abstract_actor* dest_;
+};
+
+template <class... Ts>
+class inject_clause {
+public:
+  explicit inject_clause(caf::scheduler::test_coordinator& sched)
+    : sched_(sched), dest_(nullptr) {
+    // nop
+  }
+
+  inject_clause(inject_clause&& other) = default;
+
+  template <class Handle>
+  inject_clause& from(const Handle& whom) {
+    src_ = caf::actor_cast<caf::strong_actor_ptr>(whom);
+    return *this;
+  }
+
+  template <class Handle>
+  inject_clause& to(const Handle& whom) {
+    dest_ = caf::actor_cast<caf::strong_actor_ptr>(whom);
+    return *this;
+  }
+
+  void with(Ts... xs) {
+    if (src_ == nullptr)
+      CAF_FAIL("missing .from() in inject() statement");
+    if (dest_ == nullptr)
+      CAF_FAIL("missing .to() in inject() statement");
+    caf::send_as(caf::actor_cast<caf::actor>(src_),
+                 caf::actor_cast<caf::actor>(dest_), xs...);
+    CAF_REQUIRE(sched_.prioritize(dest_));
+    auto dest_ptr = &sched_.next_job<caf::abstract_actor>();
+    auto ptr = dest_ptr->peek_at_next_mailbox_element();
+    CAF_REQUIRE(ptr != nullptr);
+    CAF_REQUIRE_EQUAL(ptr->sender, src_);
+    // TODO: replace this workaround with the make_tuple() line when dropping
+    //       support for GCC 4.8.
+    std::tuple<Ts...> tmp{std::move(xs)...};
+    using namespace caf::detail;
+    elementwise_compare_inspector<decltype(tmp)> inspector{tmp};
+    auto ys = extract<Ts...>(dest_);
+    auto ys_indices = get_indices(ys);
+    CAF_REQUIRE(apply_args(inspector, ys_indices, ys));
+    run_once();
+  }
+
+protected:
+  void run_once() {
+    auto ptr = caf::actor_cast<caf::abstract_actor*>(dest_);
+    auto dptr = dynamic_cast<caf::blocking_actor*>(ptr);
+    if (dptr == nullptr)
+      sched_.run_once();
+    else
+      dptr->dequeue(); // Drop message.
+  }
+
+  caf::scheduler::test_coordinator& sched_;
+  caf::strong_actor_ptr src_;
+  caf::strong_actor_ptr dest_;
 };
 
 template <class... Ts>
 class allow_clause {
 public:
-  allow_clause(caf::scheduler::test_coordinator& sched)
-      : sched_(sched),
-        dest_(nullptr) {
+  explicit allow_clause(caf::scheduler::test_coordinator& sched)
+    : sched_(sched), dest_(nullptr) {
     peek_ = [=] {
       if (dest_ != nullptr)
         return try_extract<Ts...>(dest_) != caf::none;
@@ -402,19 +519,16 @@ public:
   template <class Handle>
   allow_clause& to(const Handle& whom) {
     if (sched_.prioritize(whom))
-      dest_ = &sched_.next_job<caf::scheduled_actor>();
-    return *this;
-  }
-
-  allow_clause& to(const caf::scoped_actor& whom) {
-    dest_ = whom.ptr();
+      dest_ = &sched_.next_job<caf::abstract_actor>();
     return *this;
   }
 
   template <class... Us>
   void with(Us&&... xs) {
+    // TODO: replace this workaround with make_tuple() when dropping support
+    //       for GCC 4.8.
+    std::tuple<typename std::decay<Us>::type...> tmp{std::forward<Us>(xs)...};
     // TODO: move tmp into lambda when switching to C++14
-    auto tmp = std::make_tuple(std::forward<Us>(xs)...);
     peek_ = [=] {
       using namespace caf::detail;
       elementwise_compare_inspector<decltype(tmp)> inspector{tmp};
@@ -448,8 +562,8 @@ protected:
 
   caf::scheduler::test_coordinator& sched_;
   caf::strong_actor_ptr src_;
-  caf::local_actor* dest_;
-  std::function<bool ()> peek_;
+  caf::abstract_actor* dest_;
+  std::function<bool()> peek_;
 };
 
 template <class... Ts>
@@ -457,14 +571,14 @@ class disallow_clause {
 public:
   disallow_clause() {
     check_ = [=] {
-      auto ptr = next_mailbox_element(dest_);
+      auto ptr = dest_->peek_at_next_mailbox_element();
       if (ptr == nullptr)
         return;
       if (src_ != nullptr && ptr->sender != src_)
         return;
       auto res = try_extract<Ts...>(dest_);
       if (res != caf::none)
-        CAF_FAIL("received disallowed message: " << CAF_ARG(*res));
+        CAF_FAIL("received disallowed message: " << caf::deep_to_string(*ptr));
     };
   }
 
@@ -491,10 +605,12 @@ public:
 
   template <class... Us>
   void with(Us&&... xs) {
+    // TODO: replace this workaround with make_tuple() when dropping support
+    //       for GCC 4.8.
+    std::tuple<typename std::decay<Us>::type...> tmp{std::forward<Us>(xs)...};
     // TODO: move tmp into lambda when switching to C++14
-    auto tmp = std::make_tuple(std::forward<Us>(xs)...);
     check_ = [=] {
-      auto ptr = next_mailbox_element(dest_);
+      auto ptr = dest_->peek_at_next_mailbox_element();
       if (ptr == nullptr)
         return;
       if (src_ != nullptr && ptr->sender != src_)
@@ -514,7 +630,7 @@ public:
 protected:
   caf_handle src_;
   caf_handle dest_;
-  std::function<void ()> check_;
+  std::function<void()> check_;
 };
 
 template <class... Ts>
@@ -523,9 +639,10 @@ struct test_coordinator_fixture_fetch_helper {
   std::tuple<Ts...>
   operator()(caf::response_handle<Self, Interface, true>& from) const {
     std::tuple<Ts...> result;
-    from.receive(
-      [&](Ts&... xs) { result = std::make_tuple(std::move(xs)...); },
-      [&](caf::error& err) { FAIL(from.self()->system().render(err)); });
+    from.receive([&](Ts&... xs) { result = std::make_tuple(std::move(xs)...); },
+                 [&](caf::error& err) {
+                   FAIL(from.self()->system().render(err));
+                 });
     return result;
   }
 };
@@ -535,9 +652,10 @@ struct test_coordinator_fixture_fetch_helper<T> {
   template <class Self, class Interface>
   T operator()(caf::response_handle<Self, Interface, true>& from) const {
     T result;
-    from.receive(
-      [&](T& x) { result = std::move(x); },
-      [&](caf::error& err) { FAIL(from.self()->system().render(err)); });
+    from.receive([&](T& x) { result = std::move(x); },
+                 [&](caf::error& err) {
+                   FAIL(from.self()->system().render(err));
+                 });
     return result;
   }
 };
@@ -551,23 +669,35 @@ public:
   /// A deterministic scheduler type.
   using scheduler_type = caf::scheduler::test_coordinator;
 
+  /// A buffer for serializing or deserializing objects.
+  using byte_buffer = std::vector<char>;
+
   // -- constructors, destructors, and assignment operators --------------------
 
+  static Config& init_config(Config& cfg) {
+    cfg.set("logger.file-verbosity", caf::atom("quiet"));
+    if (auto err = cfg.parse(caf::test::engine::argc(),
+                             caf::test::engine::argv()))
+      CAF_FAIL("failed to parse config: " << to_string(err));
+    cfg.set("scheduler.policy", caf::atom("testing"));
+    cfg.set("logger.inline-output", true);
+    cfg.set("middleman.network-backend", caf::atom("testing"));
+    cfg.set("middleman.manual-multiplexing", true);
+    cfg.set("middleman.workers", size_t{0});
+    return cfg;
+  }
+
   template <class... Ts>
   explicit test_coordinator_fixture(Ts&&... xs)
-      : cfg(std::forward<Ts>(xs)...),
-        sys(cfg.parse(caf::test::engine::argc(), caf::test::engine::argv())
-               .set("scheduler.policy", caf::atom("testing"))
-               .set("logger.inline-output", true)
-               .set("middleman.network-backend", caf::atom("testing"))),
-        self(sys, true),
-        sched(dynamic_cast<scheduler_type&>(sys.scheduler())) {
+    : cfg(std::forward<Ts>(xs)...),
+      sys(init_config(cfg)),
+      self(sys, true),
+      sched(dynamic_cast<scheduler_type&>(sys.scheduler())) {
     // Configure the clock to measure each batch item with 1us.
     sched.clock().time_per_unit.emplace(caf::atom("batch"),
                                         caf::timespan{1000});
     // Make sure the current time isn't 0.
     sched.clock().current_time += std::chrono::hours(1);
-    credit_round_interval = cfg.stream_credit_round_interval;
   }
 
   virtual ~test_coordinator_fixture() {
@@ -630,6 +760,13 @@ public:
     return run_until([] { return false; });
   }
 
+  /// Consume ones message or triggers the next timeout.
+  /// @returns `true` if a message was consumed or timeouts was triggered,
+  ///          `false` otherwise.
+  bool run_once() {
+    return run_until([] { return true; }) > 0;
+  }
+
   /// Consume messages and trigger timeouts until `pred` becomes `true` or
   /// until no activity remains.
   /// @returns The total number of events, i.e., messages consumed and
@@ -702,6 +839,29 @@ public:
     return dynamic_cast<T&>(*ptr);
   }
 
+  template <class... Ts>
+  byte_buffer serialize(const Ts&... xs) {
+    byte_buffer buf;
+    caf::binary_serializer sink{sys, buf};
+    if (auto err = sink(xs...))
+      CAF_FAIL("serialization failed: " << sys.render(err));
+    return buf;
+  }
+
+  template <class... Ts>
+  void deserialize(const byte_buffer& buf, Ts&... xs) {
+    caf::binary_deserializer source{sys, buf};
+    if (auto err = source(xs...))
+      CAF_FAIL("deserialization failed: " << sys.render(err));
+  }
+
+  template <class T>
+  T roundtrip(const T& x) {
+    T result;
+    deserialize(serialize(x), result);
+    return result;
+  }
+
   // -- member variables -------------------------------------------------------
 
   /// The user-generated system config.
@@ -715,36 +875,8 @@ public:
 
   /// Deterministic scheduler.
   scheduler_type& sched;
-
-  // -- deprecated functionality -----------------------------------------------
-
-  void run_exhaustively() CAF_DEPRECATED_MSG("use run() instead");
-
-  void run_exhaustively_until(std::function<bool()> f)
-    CAF_DEPRECATED_MSG("use run_until() instead");
-
-  void loop_after_next_enqueue()
-    CAF_DEPRECATED_MSG("use run_after_next_ready_event() instead");
-
-  caf::timespan credit_round_interval CAF_DEPRECATED;
 };
 
-template <class Config>
-void test_coordinator_fixture<Config>::run_exhaustively() {
-  run();
-}
-
-template <class Config>
-void test_coordinator_fixture<Config>::run_exhaustively_until(
-  std::function<bool()> f) {
-  run_until(std::move(f));
-}
-
-template <class Config>
-void test_coordinator_fixture<Config>::loop_after_next_enqueue() {
-  sched.after_next_enqueue([=] { run(); });
-}
-
 /// Unboxes an expected value or fails the test if it doesn't exist.
 template <class T>
 T unbox(caf::expected<T> x) {
@@ -761,6 +893,14 @@ T unbox(caf::optional<T> x) {
   return std::move(*x);
 }
 
+/// Unboxes an optional value or fails the test if it doesn't exist.
+template <class T>
+T unbox(T* x) {
+  if (x == nullptr)
+    CAF_FAIL("x == nullptr");
+  return *x;
+}
+
 /// Expands to its argument.
 #define CAF_EXPAND(x) x
 
@@ -774,6 +914,12 @@ T unbox(caf::optional<T> x) {
     expect_clause<CAF_EXPAND(CAF_DSL_LIST types)>{sched}.fields;               \
   } while (false)
 
+#define inject(types, fields)                                                  \
+  do {                                                                         \
+    CAF_MESSAGE("inject" << #types << "." << #fields);                         \
+    inject_clause<CAF_EXPAND(CAF_DSL_LIST types)>{sched}.fields;               \
+  } while (false)
+
 /// Convenience macro for defining allow clauses.
 #define allow(types, fields)                                                   \
   ([&] {                                                                       \
