@@ -90,13 +90,13 @@
 #define SIN_LEN
 
 /* The size of `long int', as computed by sizeof. */
-#define SIZEOF_LONG_INT 4
+#define SIZEOF_LONG_INT 8
 
 /* The size of `long long', as computed by sizeof. */
 #define SIZEOF_LONG_LONG 8
 
 /* The size of `void *', as computed by sizeof. */
-#define SIZEOF_VOID_P 4
+#define SIZEOF_VOID_P 8
 
 /* should we declare syslog() and openlog() */
 /* #undef SYSLOG_INT */
@@ -129,15 +129,15 @@
 /* #undef USE_ELASTICSEARCH */
 
 /* Version number of package */
-#define VERSION "3.1.2"
+#define VERSION "3.1.5"
 
 // Zeek version number.
 // This is the result of (major * 10000 + minor * 100 + patch)
 // For example, 3.1.2 becomes 30102.
-#define ZEEK_VERSION_NUMBER 30102
+#define ZEEK_VERSION_NUMBER 30105
 
 /* whether words are stored with the most significant byte first */
-#define WORDS_BIGENDIAN
+/* #undef WORDS_BIGENDIAN */
 
 /* whether htonll/ntohll is defined in <arpa/inet.h> */
 /* #undef HAVE_BYTEORDER_64 */
@@ -229,7 +229,7 @@
 #define HAVE_IP6_EXT
 
 /* String with host architecture (e.g., "linux-x86_64") */
-#define HOST_ARCHITECTURE "openbsd-macppc"
+#define HOST_ARCHITECTURE "openbsd-loongson"
 
 /* String with extension of dynamic libraries (e.g., ".so") */
 #define DYNAMIC_PLUGIN_SUFFIX ".so"
@@ -240,7 +240,7 @@
 #endif
 
 /* A C function that has the Zeek version encoded into its name. */
-#define BRO_VERSION_FUNCTION bro_version_3_1_2_plugin_7
+#define BRO_VERSION_FUNCTION bro_version_3_1_5_plugin_7
 #ifdef __cplusplus
 extern "C" {
 #endif
